/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import com.funambol.storage.ComplexSerializer;
import com.funambol.storage.Serializable;
import com.funambol.util.StringUtil;
import com.funambol.util.Log;

import java.util.Vector;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * A class describing the RFC 2822 address specification as per  3.4
 */
public class Address implements Serializable {

    // -------------------------------------------------------------- Constants

    /** Address is used in a FROM recipient */
    public static final int FROM = 1;
    
    /** Address is used in a TO recipient */
    public static final int TO = 2;
    
    /** Address is used in a CC recipient */
    public static final int CC = 3;
    
    /** Address is used in a BCC recipient */
    public static final int BCC = 4;

    /** Address is used in a REPLYTO recipient */
    public static final int REPLYTO = 5;


    private static final char QUOTE  = '\"';
    private static final char ESCAPE = '\\';
    private static final char COMMA  = ',';
    private static final char LT     = '<';
    private static final char GT     = '>';

    // ----------------------------------------------------------- Private Data
    
    /** The type of the address */
    private int type;
    /** The optional display name (part 1 of the mailbox) */
    private String name;
    /** The email address (part 2 of the mailbox) */
    private String email;
    
    // ----------------------------------------------------------- Constructors

    /**
     * Default Constructor
     */
    public Address() {
    }
    
    /**
     * Initializes a new instance of Address, parsing <code>address<code/>.
     * If <code>address<code/> contains more than one email address,
     * only the first one is used and the others are ignored.
     * Use Adress.parse to get an array of Address from an address list.
     *
     * @param type is the type of this address
     * @param address is the String representation of the address to be parsed
     * @throws MailException when some error occurs parsing address 
     */
    public Address(int type, String address) throws MailException {
        this.type = type;
        Vector vnames = new Vector();
        Vector addresses = new Vector();

        parseAddrList(address, vnames, addresses);
        if(addresses.size() == 0) {
            Log.error("Address: invalid address '"+address+"'");
            throw new MailException(
                        MailException.INVALID_ADDRESS,
                        "Invalid empty address");
        }
        this.name = (String)vnames.elementAt(0);
        this.email = checkAddr((String)addresses.elementAt(0));
    }
    
    
    /**
     * Initializes a new instance of Address
     * @param type is the type of address
     * @param name is the name of the Address
     * @param address is the email of this address  
     */
    public Address(int type, String name, String address) throws MailException {
        this.type = type;
        this.name = name;
        this.email= checkAddr(address);
    }
    
    
    // --------------------------------------------------------- Public Methods
    
    /**
     * Accessor method to get the type of this address
     * @return int representation of this address' type
     */
    public int getType() {
        return type;
    }
    
    /**
     * Accessor method to get this address' visible name representation:
     * @return this address name if it is not null, this address email otherwise
     */
    public String getVisibleName() {
        return this.name != null ? this.name : this.email;
    }
    
    /**
     * Accessor method to get this address name field
     * @return String representation of this address name field
     */
    public String getName() {
        return name;
    }
    
    /**
     * Accessor method to get this address email field
     * @return String representation of this address email field
     */
    public String getEmail() {
        return email;
    }
    
    /**
     * Accessor method to set the type of this address
     * @param type is the type to be set for this address
     */
    public void setType(int type) {
        this.type = type;
    }
    
    /**
     * Accessor method to set the name of this address
     * @param name is the name to be set for this address
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * Accessor method to set this address' email
     * @param email is the email string value to be set
     * @throws MailException if some error occurs while check address' email 
     */
    public void setEmail(String email) throws MailException {
        this.email = checkAddr(email);
    }
    
    
    /**
     * Parses an address list and return an array of <code>Address</code>
     * objects, each representing a recipient.
     *
     * Use: Address [] list = Address.parse( Address.TO, "John
     * Doe&lt;john.doe@mail.com&gt;"
     *
     * @param addresslist
     *            the comma separated list of addresses
     * @return an array of <code>Address</code> containing the result, or an
     * empty array if addresslist is empty
     *
     * @throws MailException
     *             if the address is malformed
     *
     */
    public static Address[] parse(int type, String addresslist)
    throws MailException {

        if (addresslist == null || addresslist.trim().length() == 0) {
            return new Address[0];
        }

        // TODO: handle groups.
        if (StringUtil.equalsIgnoreCase(addresslist.trim(),
                                        "undisclosed-recipients:;")) {
            return new Address[0];
        }

        Vector vnames = new Vector();
        Vector addresses = new Vector();

        parseAddrList(addresslist, vnames, addresses);

        int len = addresses.size();
        Address[] ret = new Address[len];

        for (int i = 0; i < len; i++) {
            ret[i] = new Address(type, 
                                 (String)vnames.elementAt(i),
                                 (String)addresses.elementAt(i));
        }
        return ret;
    }


    /**
     * Overloads toString() Object' method to return this address representation
     * in RFC2822 format
     * @return the string representation of the Address
     */
    public String toString() {
        // If a name is defined, compose the '"name" <address>' form
        if (name != null) {
            StringBuffer ret = new StringBuffer("");
            
            ret.append("\"").append(name).append("\"").
                    append(" <").append(email).append(">");
            
            return ret.toString();
        }
        
        // Otherwise return the address
        return email;
    }
    
    /**
     * Serializes this instances state into a DataOutputStream
     * @param out is the DataOutputStream to write on
     * @throws IOException if some error occurs while writing on the stream
     */
    public void serialize(DataOutputStream out) throws IOException {
        ComplexSerializer.writeField(out, name);
        ComplexSerializer.writeField(out, email);
        out.writeInt(type);
    }
    
    /**
     * Deserializes this instances state from a given DataInputStream
     * @param in is the DataInputStream to read from
     * @throws IOException if some error occurs while reading the given stream
     */
    public void deserialize(DataInputStream in) throws IOException {
        name = ComplexSerializer.readField(in);
        email = ComplexSerializer.readField(in);
        type = in.readInt();
    }
    
    // -------------------------------------------------------- Private Methods

    /*
     * Parse an address list, returning the visible names and addresses in
     * the given Vectors.
     *
     * The string is parsed char by char to find special characters:
     *  ", \, <, >
     *
     * These characters change the state of the parser. The states are 4:
     *
     * Normal: at the beginning, characters are appended to the 'name' buffer.
     *         Special characters are allowed. If a '<' is found, the name is
     *         used for the visible name and the address is processed in the
     *         state 'Inside address'. Otherwise, the characters processed in
     *         this state are actually the address
     * Inside quotes: the other characters have no special meaning, ends with
     *                another quote
     * After escape: the single character has no special meaning, ends immediately
     *
     * Inside address: the special characters are not allowed, all the others are
     *                 appended to 'addr'.
     *
     * When a comma or the end of the string is found being in the Normal state,
     * the Address is complete and the name and address are appended to the output
     * Vectors.
     *
     * @param addresslist the address list to parse
     * @param vnames (out) the visible names. If an address has no visible name, its
     *               position in the vector is null.
     * @param addresses (out) the mail addresses.
     */
    private static void parseAddrList(
                            String addresslist,
                            Vector vnames,
                            Vector addresses) throws MailException {

        if (addresslist == null) {
            throw new MailException(
                        MailException.INVALID_ADDRESS,
                        "Invalid address (null)");
        }

        final int STATE_NORMAL = 0;     // beginning of the parsing
        final int STATE_QUOTE  = 1;     // inside quotes
        final int STATE_ESCAPE = 2;     // after an ESCAPE
        final int STATE_ADDR   = 3;     // inside address

        int state = STATE_NORMAL;
        StringBuffer name = new StringBuffer();
        StringBuffer addr = new StringBuffer();

        for(int i=0, l=addresslist.length(); i<l; i++) {
            char ch = addresslist.charAt(i);
            switch(state) {
                case STATE_ESCAPE:
                    name.append(ch);
                    state = STATE_NORMAL;
                    break;

                case STATE_QUOTE:
                    if (ch == QUOTE) {
                        state = STATE_NORMAL;
                    }
                    else {
                        name.append(ch);
                    }
                    break;

                case STATE_ADDR:
                    switch(ch) {
                        case QUOTE:
                        case ESCAPE:
                        case COMMA:
                        case LT:
                            throw new
                                MailException(MailException.INVALID_ADDRESS,
                                              "Invalid address.");
                        case GT:
                            state = STATE_NORMAL;
                            break;
                        default:
                            addr.append(ch);
                    }
                    break;

                case STATE_NORMAL:
                    switch(ch) {
                        case QUOTE:
                            state = STATE_QUOTE;
                            break;
                        case ESCAPE:
                            state = STATE_ESCAPE;
                            break;
                        case LT:
                            state = STATE_ADDR;
                            break;
                        case COMMA:
                            addNames(name, addr, vnames, addresses);
                            break;
                        default:
                            name.append(ch);
                    }
                    break;
            }
        }

        addNames(name, addr, vnames, addresses);

    }

    // Add visible name and email address to the vectors used by parseAddrList.
    private static void addNames(
                            StringBuffer name,
                            StringBuffer addr,
                            Vector vnames,
                            Vector addresses) throws MailException {

        // If we had name and address, save both
        // if addr is null, name is actually the address
        String vn = name.toString().trim();
        String ad = addr.toString().trim();
        
        if (ad.length() > 0) {
            vnames.addElement((vn.length() > 0) ? vn : null);
            addresses.addElement(ad);
        }
        else if (vn.length() > 0) {
            vnames.addElement(null);
            addresses.addElement(vn);
        }
        else {
            Log.error("addnames: empty address");
            throw new MailException(MailException.INVALID_ADDRESS, "Invalid address.");
        }
        name.setLength(0);
        addr.setLength(0);
    }

    /**
     * Check if the given address is a valid address
     * @param address is the adress to be checked 
     * @throws MailException when the given address is invalid 
     */
    private String checkAddr(String address) throws MailException {
        int position=address.indexOf("@");
        if ( (position!=-1) && (position!=0) && (position!=address.length())) {

            // If the address contains trailing characters, just ignore the
            // trailer. It may happen with sender using the obsolete syntax:
            //     "user@host.domain (User Name)"
            int pos = address.indexOf(" ");
            if (pos > 0) {
                address = address.substring(0, pos);
            }

            return address;
            
        } else {
            throw new MailException(
                        MailException.INVALID_ADDRESS,
                        "checkAddr: Invalid email address: " + address);
        }
    }
}

