/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;


/**
 * Represents a mailbox folder on the device. Folders can contain
 * <code>Message</code> objects, other <code>Folder</code> objects or both.
 * This nesting capability enables a folder hierarchy for stored messages. The
 * different levels of hierarchy in a folder's full name are separated by the
 * hierarchy separator character, which you can retrieve using
 * <code>getSeparator()</code>. You can retrieve a contained folder by name
 * using <code>getFolder()</code>, or a list of contained folders by using
 * <code>list()</code>
 */
public class Folder {

    // -------------------------------------------------------------- Attributes

    /**
     * The name of the folder
     */
    private String fullname = null;

    /**
     * A store contains the list of main Folders, and each <code>Folder</code>
     * a reference to its <code>Store</code>
     */
    private Store store = null;

    /**
     * A cache of the the subfolders contained by this <code>Folder</code>
     * 
     * TODO: evaluate the weight and benefits of this cache.
     */
    private Folder[] children = null;

    /**
     * The parent Folder of this Folder
     */
    private Folder parent = null;

    /**
     * The list of the contained message Ids.
     */
    private String[] messageIDs;

    // ------------------------------------------------------------ Constructors

    /*
     * The standard constructor is private, because this class must be
     * istantiated with the name and store.
     */
    private Folder() {
    }

    /**
     * Constructs a new <code>Folder</code> providing a name for it and a
     * reference to the <code>Store</code> in which it has to be created
     * 
     * @param fullname
     *            The name for this <code>Folder</code>
     * @param store
     *            The <code>Store</code> in which the folder has to be created
     */
    public Folder(String fullname, Store store) {
        this.fullname = fullname;
        this.store = store;
    }

    /**
     * Appends the given <code>Message</code> to this <code>Folder</code>
     */
    public void appendMessage(Message msg) throws MailException {
        msg.setParent(this);
        store.saveMessage(fullname, msg);
    }

    /**
     * Gets the <code>Message</code> object corresponding to the given message
     * ID
     */
    public Message getMessage(String msgid) throws MailException {
        Message ret = store.readMessage(fullname, msgid);
        ret.setParent(this);
        return ret;
    }

    /**
     * This method returns the headers of all the messages in the
     * folder <code>path</code>.
     *
     * @return An array of <code>Message</code> with all the headers
     *          set but without content
     *
     * @throws MailException
     *             If an error occurs accessing the Store
     */
    public Message[] getMsgHeaders() throws MailException {
        Message[] ret = store.getMsgHeaders(fullname);
        for(int i=0, l=ret.length; i<l; i++) {
            ret[i].setParent(this);
        }
        return ret;
    }
    
    /**
     * Gets the <code>Message</code> object corresponding using a positional
     * index.
     * The position starts from 1 and it is always referred to the messages
     * currently present in the folder. Deleting a message causes the index
     * to be shifted. This must be taken into account by calling methods.
     * The order in which the messages are ordered is not defined, and the
     * caller must not rely on it.
     *
     * @param index the position of the message to retrieve
     * @return the requested message
     */
    public Message getMessage(int index) throws MailException {
        if(index < 1){
            throw new MailException(
                        MailException.FOLDER_ERROR,
                        "[getMessage] Invalid index: " + index);
        }
        // Index starts from 1, adapt it to array index
        index--;
        // Get the array of message IDs
        String[] messageIDs = store.getMessageIDs(fullname);
        // Check index
        if (index>messageIDs.length){
            throw new MailException(
                        MailException.FOLDER_ERROR,
                        "[getMessage] Invalid index: " + index);
        }
        return store.readMessage(fullname, messageIDs[index]);
    }

    /**
     * Deletes a <code>Message</code> from this <code>Folder</code>
     * 
     * @param msg
     *            A reference to the message to be deleted
     */
    public void deleteMessage(Message msg) throws MailException {
        store.removeMessage(fullname, msg.getMessageId());
    }

    /**
     * Deletes a <code>Message</code> from this <code>Folder</code>, using
     * the message id as key.
     *
     * @param msgid
     *            A reference to the message to be deleted
     */
    public void deleteMessage(String msgid) throws MailException {
        store.removeMessage(fullname, msgid);
    }

    /**
     * Deletes a <code>Message</code> from this <code>Folder</code>, using
     * a positional index as key.
     * 
     * The position starts from 1 and it is always referred to the messages
     * currently present in the folder. Deleting a message causes the index
     * to be shifted. This must be taken into account by calling methods.
     * The order in which the messages are ordered is not defined, and the
     * caller must not rely on it.
     * 
     * @param index
     *            The index of the message to delete
     */
    public void deleteMessage(int index) throws MailException {
        if(index < 1){
            throw new MailException(
                        MailException.FOLDER_ERROR,
                        "[getMessage] Invalid index: " + index);
        }
        // Index starts from 1, adapt it to array index
        index--;
        // Get the array of message IDs
        String[] messageIDs = store.getMessageIDs(fullname);
        // Check index
        if (index>messageIDs.length){
            throw new MailException(
                        MailException.FOLDER_ERROR,
                        "[getMessage] Invalid index: " + index);
        }
        store.removeMessage(fullname, messageIDs[index]);
    }

    /**
     * Get the total number of messages in this <code>Folder</code>
     */
    public int getMessageCount() throws MailException {
        return store.countMessages(fullname);
    }

    /**
     * Returns the parent Folder of this Folder
     * 
     * @return The Folder object this Folder is child of
     */
    public Folder getParent() throws MailException {
        if (parent == null){
            int idx = fullname.lastIndexOf('/');
            // To be a subfolder, the name must be at least "/a/b"
            if(idx > 1) {
                String parentName = fullname.substring(0, idx);
                parent = store.getFolder(parentName);
            }
        }
        return parent;
    }

    /**
     * Returns the delimiter character that separates this <code>Folder</code>'s
     * pathname from the names of immediate subfolders
     * 
     * TODO: Perhaps it is better to define in a field the character used as
     * separator and not directly here (this adopted here is the solution by
     * Sun)
     * 
     * @return The hierarchy separator character
     */
    public char getSeparator() {
        return '/';
    }

    /**
     * This method returns the list of folders under this Folder
     */
    public Folder[] list() throws MailException{
        if(children == null){
            children= store.list(fullname);
        }
        return children;
    }

    /**
     * Retrieves the <code>Store</code> containing this <code>Folder</code>
     * 
     * @return A reference to the Store containing this Folder
     */
    public Store getStore() {
        return store;
    }

    /**
     * Returns the full name of this <code>Folder</code> as passed in the
     * constructor
     * 
     * @return The string containing the full name of this <code>Folder</code>
     */
    public String getFullName() {
        return fullname;
    }

    /**
     * Returns the name of this <code>Folder</code>, taking the last part of
     * the fullname
     * 
     * @return The string containing the name of this <code>Folder</code>
     */
    public String getName() {
        return fullname.substring(
                    fullname.lastIndexOf('/') + 1, fullname.length()
                );
    }

    /**
     * Returns the requested subfolder <p>
     * 
     * @param name
     *            The name of the subfolder
     * @return The <code>Folder</code> object representing the requested
     *         subfolder
     */
    public Folder getFolder(String name) throws MailException {
        String path = fullname+'/'+name;
        Folder ret = null;
        if (children != null){
            ret = searchChild(path);
        }
        else{
            ret = store.getFolder(path);
        }
        return ret;
    }

    //---------------------------------------------------------- Private methods

    // Search a child inside the Folder cache.
    private Folder searchChild(String path){
        for(int i=0; i<children.length; i++){
            if(children[i].getName().equals(path)){
                return children[i];
            }
        }
        return null;
    }

}

