/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import java.util.Vector;
import java.util.Enumeration;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.funambol.storage.Serializable;
import com.funambol.storage.ComplexSerializer;

/**
 * Represents a container for <code>BodyPart</code>s in multi-part
 * <code>Message</code>s as per RFC 2046
 */
public class Multipart implements Serializable {

    // ------------------------------------------------------------- Attributes

    /** The list of the BodyParts contained in this Multipart */
    private Vector parts;

    /** The Message or BodyPart eventually containing this Multipart. */
    private Part container;

    // ----------------------------------------------------------- Constructors

    /**
     * Constructs a Multipart object of the given content type.
     * 
     * A unique boundary string is generated and this string is setup as the
     * "boundary" string to separate BodyParts in the formatted message. <p>
     * 
     * BodyParts can be added later
     * 
     */
    public Multipart() {

        // Init the BodyPart list with an empty vector
        parts = new Vector(0, 5);
        
        container = null;
    }


    // --------------------------------------------------------- Public Methods

    /**
     * Return the container object for this Multipart.
     * 
     * @return A <code>Message</code> or a <code>BodyPart</code>
     */
    public Part getContainer() {
        return container;
    }


    /**
     * Set the container object for this Multipart.
     * 
     * Normally called by the Message or BodyPart setContent(Multipart) method
     * 
     * @param container
     */
    public void setContainer(Part container) {
        this.container = container;
    }


    /**
     * Insert the part at the end of the MultiPart.
     * 
     * @param part
     *            The BodyPart to be added to this Multipart container
     */
    public void addBodyPart(BodyPart part) {
        // Set the container reference to this Multipart
        part.setContainer(this);
        parts.addElement(part);
    }


    /**
     * Insert the part at the specified index.
     * 
     * @param part
     *            The BodyPart to be added to this Multipart container
     * @param index
     *            The position
     */
    public void addBodyPart(BodyPart part, int index) {
        // Set the container reference to this Multipart
        part.setContainer(this);
        parts.insertElementAt(part, index);
    }


    /**
     * Get the BodyPart at the specified index.
     * 
     * @param index
     *            the index of the BodyPart
     * @return the requested BodyPart
     */
    public BodyPart getBodyPart(int index) {
        return (BodyPart)parts.elementAt(index);
    }


    /**
     * Removes the BodyPart at the specified index.
     * 
     * @param index
     *            the index of the BodyPart
     */
    public void removeBodyPart(int index) {
        // Invalid the container reference
        ((BodyPart)parts.elementAt(index)).setContainer(null);
        // Remove the BodyPart from the list
        parts.removeElementAt(index);
    }


    /**
     * Retrieves the number of body parts contained in this multi-part
     */
    public int getCount() {
        return parts.size();
    }


    // ---------------------------------- interface Serializable implementation

    public void serialize(DataOutputStream dout) throws IOException {

        ComplexSerializer.serializeVector(dout, parts);

    }


    public void deserialize(DataInputStream din) throws IOException {

        parts = ComplexSerializer.deserializeVector(din);

        // Restore the container reference for all the BodyParts
        for (Enumeration e = parts.elements(); e.hasMoreElements();) {
            BodyPart p = (BodyPart)e.nextElement();
            p.setContainer(this);
        }
    }

}
