/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import java.util.Vector;

/**
 * Represents a message store and its access protocol, for storing and
 * retrieving messages on the device
 */
public interface Store {

    /* Default Name for the Inbox folders*/
    public static final String INBOX  = "Inbox";
    /* Default Name for the Outbox folders*/
    public static final String OUTBOX = "Outbox";
    /* Default Name for the Draft folders*/
    public static final String DRAFTS = "Drafts";
    /* Default Name for the sent folders*/
    public static final String SENT   = "Sent";
    // Max number of messages that can be stored
    //public static int DEFAULT_MAX_MESSAGE_NUMBER = 100;

    
    /**
     * Initialize the message store, creating the main folders.
     *
     * @param reset if true, erase and re-create all the main folders.
     */
    public void init(boolean reset) throws MailException ;

    /**
     * Retrieves a list of folders directly under this Store
     * 
     * @return an array of folders, or <code>null</code> if the store is empty.
     */
    public Folder[] list();

    /**
     * This method returns the list of the folders whose path
     * starts with 'path' and are direct subfolders of it.
     *
     * @param path the path of the parent folder
     * @return an array of folders, or <code>null</code> if
     *          <code>path</code> has no subfolders.
     *
     * @throws StoreException
     *             If the path is not valid
     */
    public Folder[] list(String path) throws MailException;

    /**
     * Creates a Folder in this Store with the provided path
     * 
     * @param path
     *             the full path of the folder in the Store.     
     * @throws StoreException
     *             If an error occurs on the store (e.g. no space left)
     */
    public Folder createFolder(String path) throws MailException;

    /**
     * Removes a folder from the record store. 
     * 
     * @param path
     *             The full pathname in the Store
     * @return <code>true</code> if the folder has been actually deleted,
     *         <code>false</code> if the folder did not exist.
     * @throws StoreException
     *             if an error occurs on the store 
     */
    public boolean removeFolder(String path) throws MailException ;
     
    /**
     * Retrieves a Folder by name
     * 
     * @param path
     *            The path to the Folder in the device's file system
     * @return The Folder object
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public Folder getFolder(String path) throws MailException;

    /**
     * Retrieves a Folder in this Store that matches the provided substring
     * 
     * @param path
     *            The partial path in the device's file system to the searched
     *            Folder
     * @return A list of Folder objects whose path matches the provided
     *         substring, or <code>null</code> if there are no matches
     */
    public Folder[] findFolders(String path);

    /**
     * This method reads the headers of all the message from a folder.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            messages.
     * @return An array of <code>Message</code> with all the headers
     *          set but without content
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public Message[] getMsgHeaders(String path) throws MailException;

    /**
     * This method returns the messageIDs of all the messages in the
     * folder <code>path</code>.
     *
     * @param path the complete path of the folder
     * @return an array of String, or <code>null</code> if the folder is
     *         empty.
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public String[] getMessageIDs(String path) throws MailException;

    /**
     * This method reads a message from the Store, using the message Id.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param msgid
     *            A string representing the unique message ID 
     * @return The <code>Message</code> corresponding to the passed path and
     *         message ID from this <code>Store</code>
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    Message readMessage(String path, String msgid) throws MailException;

    /**
     * This method saves a message in the Store
     *
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param msg
     *            The <code>Message</code> to save.
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    void saveMessage(String path, Message msg) throws MailException;

    /**
     * This method removes a Message from the Store, using message ID
     * as index.
     *
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param messageID
     *            A string representing the unique message ID 
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    void removeMessage(String path, String messageID) throws MailException;

    /**
     * This method returns the message count in this Store.
     *
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     *
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    int countMessages(String path) throws MailException;
}

