/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import com.funambol.util.StringUtil;
import java.util.Date;

import jmunit.framework.cldc10.AssertionFailedException;
import jmunit.framework.cldc10.TestCase;

import com.funambol.util.Log;
import com.funambol.util.MailDateFormatter;

/**
 * A JMUnit test class to check the functionality of the MIME formatter.
 * Implementing the TestMessage interface this class has direct access to all
 * fields defined in that interface
 */
public class MIMEFormatterTest extends TestCase implements TestMessage {

    /**
     * The message object to catch the data from that are to be formatted into a
     * RFC 2822 conform text
     */
    Message message = null;

    /**
     * A store (located on the device's RMS)
     */
    RMSStore store = null;

    /**
     * A folder to hold the message
     */
    Folder folder = null;

    /**
     * Today's date
     */
    Date date = null;

    /**
     * The message's content
     */
    String content = null;

    /**
     * Carriage Return + Line Feed
     */
    private static final String CRLF = "\r\n";


    /**
     * Tests the content of the "From:" header attribute in the formatted
     * message: it has to match the original one contained in the test email
     * message stored in a string constant in the {@link TestMessage} interface
     * 
     * @throws AssertionFailedException
     */
    public void testtoString() throws AssertionFailedException {

        // sysout instead of Log.trace() because of output to Antenna console
        System.out.println("toString");

        // here only the first (and unique) item is retrieved
        MIMEFormatter formatter = null;
        try {
            formatter = new MIMEFormatter(folder.getMessage(1));
        } catch (MailException e) {
            Log.error("Error by retrieving a Message from the Folder:");
            e.printStackTrace();
        }

        String dateRfc2822 = MailDateFormatter.dateToRfc2822(date);
        String expectedResult =
            "Date: " + dateRfc2822 + CRLF +
            "From: giuseppe.monticelli@gmail.com" + CRLF +
            "User-Agent: " + CRLF + // empty string
            "MIME-Version: 1.0" + CRLF +
            "To: monty@funambol.com" + CRLF +
            "Subject: An experiment with the SyncML" + CRLF +
            "Message-ID: ---Test Msg ID ---" + CRLF +
            "Content-Type: text/plain" + CRLF +
            "Content-Transfer-Encoding: " + CRLF + // empty string
            CRLF +
            content;

        // String coming from the formatter
        String result = formatter.toString();

        System.out.println("Expected:\n\n" + expectedResult);
        System.out.println("Result:\n\n" + result);

        /*
         * The string coming from the Message object must be contained in the
         * formatted message
         */
        //assertTrue(result.equalsIgnoreCase(expectedResult));
        assertTrue(StringUtil.equalsIgnoreCase(result, expectedResult));
    }


    /**
     * The constructor. Please refer to the JMUnit documentation
     */
    public MIMEFormatterTest() {
        super(1, "MIMEFormatterTest");
        Log.setLogLevel(Log.DEBUG);
    }


    /**
     * Prepares the conditions to make the tests. A test message is parsed by
     * the {@link MIMEParser} and a {@link} Message} object is built. A new
     * {@link Folder} in the RMS store is created and the message is appended to
     * it
     * 
     * @see jmunit.framework.cldc11.TestCase#setUp()
     */
    public void setUp() throws Exception {
        Address address = new Address(Address.TO, "monty@funambol.com");
        Address[] list = { address };
        content = "Hello, Fun!";
        date = new Date();

        message = new Message();
        message.addRecipients(list);
        message.addHeader(Message.FROM, "giuseppe.monticelli@gmail.com");
        message.addHeader(Message.SUBJECT, "An experiment with the SyncML");
        message.setSentDate(date);
        message.setMessageId("---Test Msg ID ---");
        message.setContent(content);

        store = new RMSStore();
        folder = store.createFolder("inbox");
        folder.appendMessage(message);
    }


    /**
     * @see jmunit.framework.cldc11.TestCase#tearDown()
     */
    public void tearDown() {
        // not used
    }


    /**
     * @see jmunit.framework.cldc11.TestCase#test(int)
     */
    public void test(int testNumber) throws Throwable {
        switch (testNumber) {
            case 0:
                testtoString();
                break;
            default:
                break;
        }
    }

}
